from typing import List, Optional

import pytest

from sdkite.http.utils import urljoin, urlsjoin


@pytest.mark.parametrize(
    "base, url, expected",
    [
        (None, None, None),
        ("", "", None),
        ("https://perdu.com", None, "https://perdu.com"),
        ("https://perdu.com/", None, "https://perdu.com/"),
        ("https://perdu.com", "", "https://perdu.com"),
        ("https://perdu.com/", "", "https://perdu.com/"),
        (None, "https://perdu.com", "https://perdu.com"),
        (None, "https://perdu.com/", "https://perdu.com/"),
        ("", "https://perdu.com", "https://perdu.com"),
        ("", "https://perdu.com/", "https://perdu.com/"),
        ("https://perdu.com", "https://perdu.org", "https://perdu.org"),
        ("https://perdu.com", "https://perdu.org/", "https://perdu.org/"),
        ("https://perdu.com/", "https://perdu.org", "https://perdu.org"),
        ("https://perdu.com/", "https://perdu.org/", "https://perdu.org/"),
        ("https://perdu.com", "abc/def", "https://perdu.com/abc/def"),
        ("https://perdu.com/", "abc/def", "https://perdu.com/abc/def"),
        ("https://perdu.com", "/abc/def", "https://perdu.com/abc/def"),
        ("https://perdu.com/", "/abc/def", "https://perdu.com/abc/def"),
        ("https://perdu.com/abc/def/ghi", "../jkl", "https://perdu.com/abc/def/jkl"),
        ("https://perdu.com/abc/def/ghi", "/jkl", "https://perdu.com/jkl"),
    ],
)
def test_urljoin(
    base: Optional[str], url: Optional[str], expected: Optional[str]
) -> None:
    assert urljoin(base, url) == expected


@pytest.mark.parametrize(
    "urls, expected",
    [
        (
            [],
            None,
        ),
        (
            [None, None, None],
            None,
        ),
        (
            ["https://example.com"],
            "https://example.com",
        ),
        (
            [None, "https://example.com"],
            "https://example.com",
        ),
        (
            ["https://example.com", None],
            "https://example.com",
        ),
        (
            [None, "https://example.com", None],
            "https://example.com",
        ),
        (
            ["https://example.com", "abc", "def"],
            "https://example.com/abc/def",
        ),
        (
            ["https://example.com/", "abc/", "def/"],
            "https://example.com/abc/def/",
        ),
        (
            ["https://example.com", "abc", "def", "../ghi"],
            "https://example.com/abc/ghi",
        ),
        (
            ["https://example.com", "abc", "def", "/ghi"],
            "https://example.com/ghi",
        ),
        (
            ["https://example.com", "abc", "def", "https://example.org/", "jkl"],
            "https://example.org/jkl",
        ),
    ],
)
def test_urlsjoin(urls: List[Optional[str]], expected: Optional[str]) -> None:
    assert urlsjoin(urls) == expected



from io import SEEK_END, StringIO
from typing import TYPE_CHECKING, Any

from sdkite import AdapterSpec, Client

if TYPE_CHECKING:
    from typing_extensions import assert_type
else:
    # no need to have typing_extensions installed
    def assert_type(val: Any, _: Any) -> Any:
        return val


class ValueSpec(AdapterSpec[StringIO]):
    def _create_adapter(self) -> StringIO:
        return StringIO()  # not an Adapter instance


class ReadClient(Client):
    value = ValueSpec()

    def normal(self) -> str:
        self.value.seek(0)
        return self.value.read()

    def upper(self) -> str:
        return self.normal().upper()


class WriteClient(Client):
    value = ValueSpec()

    def erase(self) -> None:
        self.value.seek(0)
        self.value.truncate()

    def append(self, suffix: str) -> None:
        self.value.seek(0, SEEK_END)
        self.value.write(suffix)


class RootClient(Client):
    value = ValueSpec()

    read: ReadClient
    write: WriteClient


def test_typing() -> None:
    client = RootClient()

    assert isinstance(assert_type(client, RootClient), RootClient)
    assert isinstance(assert_type(client.read, ReadClient), ReadClient)

    assert isinstance(assert_type(client.value, StringIO), StringIO)
    assert isinstance(assert_type(client.read.value, StringIO), StringIO)

    assert isinstance(assert_type(RootClient.value, ValueSpec), ValueSpec)


def test_complete() -> None:
    client = RootClient()
    assert client.read.normal() == ""
    assert client.read.upper() == ""

    client.write.append("Hello")
    assert client.read.normal() == "Hello"
    assert client.read.upper() == "HELLO"

    client.write.append(", world!")
    assert client.read.normal() == "Hello, world!"
    assert client.read.upper() == "HELLO, WORLD!"

    client.write.erase()
    assert client.read.normal() == ""
    assert client.read.upper() == ""


def test_two_clients() -> None:
    # two client instances are independent
    client0 = RootClient()
    client1 = RootClient()
    client0.write.append("Hello!")
    client1.write.append("How are you?")
    assert client0.read.normal() == "Hello!"
    assert client1.read.normal() == "How are you?"


class RootClientMissingValue(Client):
    read: ReadClient
    write: WriteClient


def test_root_client_missing_value() -> None:
    client = RootClientMissingValue()
    assert client.read.normal() == ""
    client.write.append("Hello")
    assert client.read.normal() == ""  # did not change



from abc import ABC, abstractmethod
from copy import deepcopy
import sys
from typing import (
    Any,
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    get_type_hints,
    overload,
)

from sdkite.client import Client

A = TypeVar("A")


# until mypy has support for Self type
# see https://github.com/python/mypy/issues/11871
_AdapterSelf = TypeVar("_AdapterSelf", bound="Adapter")
_AdapterSpecSelf = TypeVar("_AdapterSpecSelf", bound="AdapterSpec[Any]")


class Adapter:
    _attr_name: str
    _clients: Tuple[Client, ...]

    @property
    def _adapters(self: _AdapterSelf) -> Tuple[_AdapterSelf, ...]:
        return tuple(getattr(client, self._attr_name) for client in self._clients)


def create_adapter_proxy(
    adapter: A,
    attr_name: str,
    clients: Tuple[Client, ...],
    context: Dict[str, Any],
) -> A:
    klass = type(adapter)

    # pylint: disable=unused-argument

    def init(self: Any) -> None:
        pass

    def getattribute(self: Any, name: str) -> Any:
        if name == "_attr_name":
            return attr_name
        if name == "_clients":
            return clients
        if name in context:
            return context[name]
        try:
            attr = super(type(self), self).__getattribute__(name)
        except AttributeError:
            attr = getattr(adapter, name)
        return attr

    def setattribute(self: Any, name: str, value: Any) -> None:
        if name in ("_attr_name", "_clients"):
            raise RuntimeError(f"Attribute {name} is read-only")  # pragma: no cover
        if name in context:
            context[name] = value
        else:
            setattr(adapter, name, value)

    return cast(
        A,
        type(
            f"{klass.__name__}*",
            (klass,),
            {
                "__init__": init,
                "__getattribute__": getattribute,
                "__setattr__": setattribute,
                "__doc__": getattr(klass, "__doc__", None),
            },
        )(),
    )


class AdapterSpec(Generic[A], ABC):
    _attr_name: str

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}"

    def __set_name__(self, _: Any, name: str) -> None:
        self._attr_name = name

    @overload
    def __get__(self: _AdapterSpecSelf, client: None, _: Any) -> _AdapterSpecSelf:
        ...

    @overload
    def __get__(self, client: Client, _: Any) -> A:
        ...

    def __get__(
        self: _AdapterSpecSelf, client: Optional[Client], _: Any
    ) -> Union[_AdapterSpecSelf, A]:
        if client is None:
            return self

        _attr_name_adapter = f"_adapter__{self._attr_name}"
        _attr_name_adapter_real = f"_adapter_real__{self._attr_name}"

        # in cache
        adapter: Optional[A] = getattr(client, _attr_name_adapter, None)

        if adapter is None:
            # walk up clients chain
            current_client = client
            last_descriptor_type: Optional[Type[AdapterSpec[Any]]] = None
            clients: List[Client] = []
            expect_no_more_descriptors = False
            while True:
                try:
                    descriptor: AdapterSpec[A] = getattr(
                        type(current_client), self._attr_name
                    )
                except AttributeError:
                    expect_no_more_descriptors = True
                else:
                    if expect_no_more_descriptors:
                        raise TypeError(
                            f"Client '{type(clients[-1]).__name__}' is missing"
                            f" adapter spec '{self._attr_name}'"
                        )
                    if last_descriptor_type and (
                        # pylint: disable-next=unidiomatic-typecheck
                        type(descriptor)
                        is not last_descriptor_type
                    ):
                        raise TypeError(
                            f"Client '{type(clients[-1]).__name__}' has a wrong adapter spec:"
                            f" got '{last_descriptor_type.__name__}'"
                            f" instead of '{type(descriptor).__name__}'"
                        )
                    last_descriptor_type = type(descriptor)
                    clients.append(current_client)
                if current_client._parent:
                    current_client = current_client._parent
                else:
                    if sys.version_info < (3, 10):  # pragma: no cover
                        # fix coverage issue on some Python versions
                        # see https://github.com/nedbat/coveragepy/issues/1480
                        pass
                    break
            clients.reverse()
            root_client = clients[0]

            # create real adapter if needed
            try:
                real_adapter: A = getattr(root_client, _attr_name_adapter_real)
            except AttributeError:
                real_adapter = self._create_adapter()
                setattr(root_client, _attr_name_adapter_real, real_adapter)

            # create adapter proxy
            if isinstance(real_adapter, Adapter):
                adapter_proxy = create_adapter_proxy(
                    real_adapter,
                    self._attr_name,
                    tuple(clients),
                    {
                        attr_name: deepcopy(getattr(self, attr_name))
                        for attr_name in get_type_hints(real_adapter.__class__)
                        if not attr_name.startswith("_")
                    },
                )
                adapter = cast(A, adapter_proxy)
            else:
                adapter = real_adapter

            setattr(client, _attr_name_adapter, adapter)

        return adapter

    @abstractmethod
    def _create_adapter(self) -> A:
        ...



from typing import Dict, List, Tuple, Union

import pytest

from sdkite.http import HTTPHeaderDict

keys_cases = {
    "lower": "abc",
    "title": "Abc",
    "middle": "aBc",
    "upper": "ABC",
}


@pytest.mark.parametrize(
    "key", [pytest.param(key, id=name) for name, key in keys_cases.items()]
)
def test_getitem(key: str) -> None:
    value = "aBc01!$"
    hdict = HTTPHeaderDict()
    assert repr(hdict) == r"HTTPHeaderDict{}"

    for key2 in keys_cases.values():
        with pytest.raises(KeyError) as exc_info:
            hdict[key2]  # pylint: disable=pointless-statement
        assert exc_info.value.args == (key2,)

    hdict[key] = value
    assert repr(hdict) == f"HTTPHeaderDict{{'{key}': ['{value}']}}"

    for key2 in keys_cases.values():
        assert key2 in hdict
        assert hdict[key2] == value


def test_setitem() -> None:
    hdict = HTTPHeaderDict()
    assert "abc" not in hdict

    hdict["abc"] = "012"
    assert hdict["abc"] == "012"
    assert len(hdict) == 1
    assert dict(hdict) == {"abc": "012"}
    assert repr(hdict) == r"HTTPHeaderDict{'abc': ['012']}"

    hdict["abc"] = "345"
    assert hdict["abc"] == "345"
    assert len(hdict) == 1
    assert dict(hdict) == {"abc": "345"}
    assert repr(hdict) == r"HTTPHeaderDict{'abc': ['345']}"

    hdict["aBc"] = "678"
    assert hdict["abc"] == "678"
    assert len(hdict) == 1
    assert dict(hdict) == {"aBc": "678"}
    assert repr(hdict) == r"HTTPHeaderDict{'aBc': ['678']}"


def test_add() -> None:
    hdict = HTTPHeaderDict()
    assert "abc" not in hdict

    hdict.add("abc", "012")
    assert hdict["abc"] == "012"
    assert len(hdict) == 1
    assert dict(hdict) == {"abc": "012"}
    assert repr(hdict) == r"HTTPHeaderDict{'abc': ['012']}"

    hdict.add("abc", "345")
    assert hdict["abc"] == "012, 345"
    assert len(hdict) == 1
    assert dict(hdict) == {"abc": "012, 345"}
    assert repr(hdict) == r"HTTPHeaderDict{'abc': ['012', '345']}"

    hdict.add("aBc", "678")
    assert hdict["abc"] == "012, 345, 678"
    assert len(hdict) == 1
    assert dict(hdict) == {"abc": "012, 345, 678"}
    assert repr(hdict) == r"HTTPHeaderDict{'abc': ['012', '345', '678']}"


def test_add_keep_first_header_case() -> None:
    hdict = HTTPHeaderDict()
    assert "abc" not in hdict

    hdict.add("abC", "012")
    assert hdict["abc"] == "012"
    assert dict(hdict) == {"abC": "012"}
    assert repr(hdict) == r"HTTPHeaderDict{'abC': ['012']}"

    hdict.add("abc", "345")
    assert hdict["abc"] == "012, 345"
    assert dict(hdict) == {"abC": "012, 345"}
    assert repr(hdict) == r"HTTPHeaderDict{'abC': ['012', '345']}"

    hdict.add("aBc", "678")
    assert hdict["abc"] == "012, 345, 678"
    assert dict(hdict) == {"abC": "012, 345, 678"}
    assert repr(hdict) == r"HTTPHeaderDict{'abC': ['012', '345', '678']}"


def test_delitem() -> None:
    hdict = HTTPHeaderDict()
    hdict["aBc"] = "012"
    assert dict(hdict) == {"aBc": "012"}
    assert repr(hdict) == r"HTTPHeaderDict{'aBc': ['012']}"

    del hdict["abc"]

    for key in keys_cases.values():
        assert key not in hdict
    assert not dict(hdict)
    assert repr(hdict) == r"HTTPHeaderDict{}"

    hdict["abC"] = "345"
    assert dict(hdict) == {"abC": "345"}
    assert repr(hdict) == r"HTTPHeaderDict{'abC': ['345']}"


@pytest.mark.parametrize(
    "items",
    [
        None,
        {
            "aBc": "middle",
            "ABC": "upper",
            "XyZ": "012",
            "abc": "lower",
            "Abc": "title",
        },
        [
            ("aBc", "middle"),
            ("ABC", "upper"),
            ("XyZ", "012"),
            ("abc", "lower"),
            ("Abc", "title"),
        ],
    ],
    ids=type,
)
def test_init(items: Union[None, Dict[str, str], List[Tuple[str, str]]]) -> None:
    hdict = HTTPHeaderDict(items)
    if items is None:
        assert not dict(hdict)
        assert repr(hdict) == r"HTTPHeaderDict{}"
    else:
        assert dict(hdict) == {
            "aBc": "middle, upper, lower, title",
            "XyZ": "012",
        }
        assert repr(hdict) == (
            "HTTPHeaderDict{"
            "'aBc': ['middle', 'upper', 'lower', 'title'], "
            "'XyZ': ['012']}"
        )



from pathlib import Path
import sys
from typing import List

import pytest


def pytest_collection_modifyitems(
    # pylint: disable=unused-argument
    config: pytest.Config,
    items: List[pytest.Item],
) -> None:
    tests_root = Path(__file__).parent
    for item in items:
        relative = Path(item.fspath).parent.relative_to(tests_root)
        mark = relative.parts[0]
        item.add_marker(getattr(pytest.mark, mark))


# skipping some files due to syntax not yet supported
collect_ignore: List[str] = []
if sys.version_info < (3, 8):
    collect_ignore.append("unit/test_pagination_pep570.py")



from typing import Optional

import pytest

from sdkite.http import HTTPBodyEncoding
from sdkite.http.utils import encode_request_body


@pytest.mark.parametrize(
    "body, encoding, expected_is_iterator, expected_data, expected_content_type",
    [
        # None
        (None, HTTPBodyEncoding.AUTO, False, b"", None),
        (None, HTTPBodyEncoding.NONE, False, b"", None),
        (None, HTTPBodyEncoding.JSON, False, b"null", "application/json"),
        (None, HTTPBodyEncoding.URLENCODE, False, None, None),
        (None, HTTPBodyEncoding.MULTIPART, False, None, None),
        # bytes
        (b"abc", HTTPBodyEncoding.AUTO, False, b"abc", None),
        (b"abc", HTTPBodyEncoding.NONE, False, b"abc", None),
        (b"abc", HTTPBodyEncoding.JSON, False, None, None),
        (b"abc", HTTPBodyEncoding.URLENCODE, False, None, None),
        (b"abc", HTTPBodyEncoding.MULTIPART, False, None, None),
        # str
        ("æther", HTTPBodyEncoding.AUTO, False, b"\xc3\xa6ther", None),
        ("æther", HTTPBodyEncoding.NONE, False, b"\xc3\xa6ther", None),
        ("æther", HTTPBodyEncoding.JSON, False, b'"\\u00e6ther"', "application/json"),
        ("æther", HTTPBodyEncoding.URLENCODE, False, None, None),
        ("æther", HTTPBodyEncoding.MULTIPART, False, None, None),
    ],
)
def test_auto(
    body: object,
    encoding: HTTPBodyEncoding,
    expected_is_iterator: bool,
    expected_data: Optional[bytes],
    expected_content_type: Optional[str],
) -> None:
    if expected_data is None:
        with pytest.raises(TypeError):
            encode_request_body(body, encoding)
    else:
        data, content_type = encode_request_body(body, encoding)
        assert isinstance(data, bytes) != expected_is_iterator
        if isinstance(data, bytes):
            assert data == expected_data, "Invalid data"
        else:
            assert b"".join(data) == expected_data, "Invalid data"
        assert content_type == expected_content_type, "Invalid content type"



from functools import reduce
import json
import sys
from typing import Optional, Tuple, Union
from urllib.parse import urljoin as _urljoin

from sdkite.http.model import HTTPBodyEncoding

if sys.version_info < (3, 9):  # pragma: no cover
    from typing import Iterable, Iterator
else:  # pragma: no cover
    from collections.abc import Iterable, Iterator


def urljoin(base: Optional[str], url: Optional[str]) -> Optional[str]:
    """
    Re-implementation of urljoin with some edge cases:
     - Handling of None values (note: this was the case with stdlib but undocumented)
     - Work more like os.path.join when base does not end with "/"
       e.g. __urljoin("https://example.com/foo", "bar")
            gives "https://example.com/foo/bar"
            and not "https://example.com/bar"
    """
    if base:
        if url:
            if not base.endswith("/"):
                base += "/"
            url = _urljoin(base, url)
        else:
            url = base
    elif not url:
        url = None
    return url


def urlsjoin(parts: Iterable[Optional[str]]) -> Optional[str]:
    return reduce(urljoin, parts, None)


def encode_request_body(
    body: object,
    encoding: HTTPBodyEncoding,
) -> Tuple[Union[bytes, Iterator[bytes]], Optional[str]]:
    type_error = TypeError(
        f"Cannot encode '{type(body).__name__}' body with '{encoding.name}' encoding"
    )

    # empty body
    if body is None:
        if encoding in (HTTPBodyEncoding.AUTO, HTTPBodyEncoding.NONE):
            return (b"", None)
        if encoding == HTTPBodyEncoding.JSON:
            return (b"null", "application/json")
        raise type_error

    # bytes
    if isinstance(body, bytes):
        if encoding in (HTTPBodyEncoding.AUTO, HTTPBodyEncoding.NONE):
            return (body, None)
        raise type_error

    # str
    if isinstance(body, str):
        if encoding in (HTTPBodyEncoding.AUTO, HTTPBodyEncoding.NONE):
            return (body.encode(), None)
        if encoding == HTTPBodyEncoding.JSON:
            return (json.dumps(body).encode(), "application/json")
        raise type_error

    # complex types
    raise NotImplementedError(
        "Encoding of complex body types is not supported yet"
    )  # pragma: no cover



from base64 import b64encode
import sys
from typing import List, cast

from requests_mock import Mocker

from sdkite import Client, Pagination, paginated
from sdkite.http import HTTPAdapterSpec

if sys.version_info < (3, 8):  # pragma: no cover
    from typing_extensions import TypedDict
else:  # pragma: no cover
    from typing import TypedDict


class User(TypedDict):
    name: str
    age: int


class ApiUsers(Client):
    http = HTTPAdapterSpec(url="users", headers={"X-Toto": "Abc"})

    def get(self, user_id: int) -> User:
        response = self.http.request("get", str(user_id))
        return cast(User, response.data_json)

    @paginated(page=1)
    def get_all(self, pagination: Pagination) -> List[User]:
        response = self.http.request("get", f"all/{pagination.page}")
        return cast(List[User], response.data_json)


class Api(Client):
    http = HTTPAdapterSpec(headers={"authorization": "Basic dGVzdHM6"})

    users: ApiUsers

    def __init__(self, url: str, password: str) -> None:
        super().__init__()
        self.http.url = url
        self.http.headers["authorization"] += b64encode(password.encode()).decode()


def test_http(requests_mock: Mocker) -> None:
    requests_mock.register_uri(
        "GET",
        "https://www.example.com/api/v1/users/1337",
        request_headers={"authorization": "Basic dGVzdHM6czNjcjN0"},
        content=b'{"name":"John Doe","age":42}',
    )

    requests_mock.register_uri(
        "GET",
        "https://www.example.com/api/v1/users/all/1",
        request_headers={"authorization": "Basic dGVzdHM6czNjcjN0"},
        content=b'[{"name":"John Doe","age":42},{"name":"Alice Doe","age":41}]',
    )
    requests_mock.register_uri(
        "GET",
        "https://www.example.com/api/v1/users/all/2",
        request_headers={"authorization": "Basic dGVzdHM6czNjcjN0"},
        content=b'[{"name":"Bob Doe","age":10},{"name":"Carole Doe","age":12}]',
    )
    requests_mock.register_uri(
        "GET",
        "https://www.example.com/api/v1/users/all/3",
        request_headers={"authorization": "Basic dGVzdHM6czNjcjN0"},
        content=b"[]",
    )

    client = Api("https://www.example.com/api/v1", "s3cr3t")

    user = client.users.get(1337)
    assert user == User(name="John Doe", age=42)

    users = list(client.users.get_all())
    assert users == [
        User(name="John Doe", age=42),
        User(name="Alice Doe", age=41),
        User(name="Bob Doe", age=10),
        User(name="Carole Doe", age=12),
    ]



# this file requires Python 3.8

from inspect import Parameter, signature
import sys
from typing import TYPE_CHECKING, Any, List, Protocol, get_type_hints

from sdkite import Pagination, paginated

if sys.version_info < (3, 9):  # pragma: no cover
    from typing import Iterator
else:  # pragma: no cover
    from collections.abc import Iterator


if TYPE_CHECKING:
    from typing_extensions import assert_type
else:
    # no need to have typing_extensions installed
    def assert_type(val: Any, _: Any) -> Any:
        return val


# this file uses positional-only parameters (PEP 570)
# which is only supported from Python 3.8

# this needs to be merged with regular test file
# when support for Python 3.7 is dropped


def test_paginated_wrapping_function_pep570() -> None:
    @paginated()
    def fct(
        pagination: Pagination,
        var_a: int,
        /,
        var_b: int,
        *var_c: int,
        var_d: int,
        var_e: int = 42,
        **var_f: int,
    ) -> List[int]:
        """Foobar"""
        if pagination.page == 0:
            return [var_a, var_b, *var_c, var_d, var_e, *var_f.values()]
        return []

    class ExpectedType(Protocol):
        def __call__(
            self,
            var_a: int,
            /,
            var_b: int,
            *var_c: int,
            var_d: int,
            var_e: int = 42,
            **var_f: int,
        ) -> Iterator[int]:
            ...

    assert_type(fct, ExpectedType)

    assert fct.__name__ == "fct", "name"
    assert fct.__doc__ == "Foobar", "doc"
    sig = signature(fct)
    assert list(sig.parameters.values()) == [
        Parameter("var_a", Parameter.POSITIONAL_ONLY, annotation=int),
        Parameter("var_b", Parameter.POSITIONAL_OR_KEYWORD, annotation=int),
        Parameter("var_c", Parameter.VAR_POSITIONAL, annotation=int),
        Parameter("var_d", Parameter.KEYWORD_ONLY, annotation=int),
        Parameter("var_e", Parameter.KEYWORD_ONLY, annotation=int, default=42),
        Parameter("var_f", Parameter.VAR_KEYWORD, annotation=int),
    ], "signature parameters"
    assert sig.return_annotation == Iterator[Any], "signature return_annotation"
    assert get_type_hints(fct) == {
        "var_a": int,
        "var_b": int,
        "var_c": int,
        "var_d": int,
        "var_e": int,
        "var_f": int,
        "return": Iterator[Any],
    }

    assert list(fct(1, 2, 3, 4, foo=5, bar=6, var_d=7)) == [1, 2, 3, 4, 7, 42, 5, 6]


def test_paginated_wrapping_method() -> None:
    class Klass:
        def __init__(self, value: int) -> None:
            self.value = value

        @paginated()
        def meth(
            self,
            pagination: Pagination,
            var_a: int,
            *var_b: int,
            var_c: int,
            var_d: int = 42,
            **var_e: int,
        ) -> List[int]:
            """Foobar"""
            if pagination.page == 0:
                return [self.value, var_a, *var_b, var_c, var_d, *var_e.values()]
            return []

    class ExpectedClassMethodType(Protocol):
        def __call__(
            self,
            obj: Klass,
            /,
            var_a: int,
            *var_b: int,
            var_c: int,
            var_d: int = 42,
            **var_e: int,
        ) -> Iterator[int]:
            ...

    assert_type(Klass.meth, ExpectedClassMethodType)



from typing import Dict, Optional, Tuple
from unittest.mock import Mock, call

import pytest

from sdkite.http import HTTPAdapter, HTTPBodyEncoding, HTTPHeaderDict, HTTPRequest


def create_adapter(
    url: Optional[str] = None,
    headers: Optional[Dict[str, str]] = None,
    parent_adapter: Optional[HTTPAdapter] = None,
) -> Tuple[HTTPAdapter, Mock, Mock]:
    # pylint: disable=protected-access
    send_request = Mock()
    adapter = HTTPAdapter(send_request)
    adapter._attr_name = "http"
    client = Mock()
    client.http = adapter
    adapter._clients = (client,)
    if parent_adapter:
        adapter._clients = parent_adapter._clients + adapter._clients
    adapter.url = url
    adapter.headers = HTTPHeaderDict(headers)
    adapter.request_interceptor = {}
    adapter.response_interceptor = {}
    return (adapter, send_request, client)


@pytest.mark.parametrize("method", ["post", "pOsT", "POST"])
def test_request_method(method: str) -> None:
    adapter, send_request, _ = create_adapter()
    response = adapter.request(method, "https://www.example.com")
    assert response == send_request.return_value
    assert send_request.call_args_list == [
        call(
            HTTPRequest(
                method="POST",
                url="https://www.example.com",
                headers=HTTPHeaderDict(),
                body=b"",
                stream_response=False,
            )
        )
    ]


@pytest.mark.parametrize(
    "base_url, url",
    [
        (None, "https://www.example.com/foo/bar"),
        ("https://www.example.com", "foo/bar"),
        ("https://www.example.com/", "foo/bar"),
        ("https://www.example.com/foo", "bar"),
        ("https://www.example.com/foo/", "bar"),
        ("https://www.example.com/foo/bar", ""),
        ("https://www.example.com/foo/bar", None),
        ("https://www.example.com", "/foo/bar"),
        ("https://www.example.com/", "/foo/bar"),
        ("https://www.example.com/abc/def", "/foo/bar"),
        ("https://www.example.com/foo/abc/def", "../../bar"),
    ],
)
def test_request_url(base_url: Optional[str], url: Optional[str]) -> None:
    adapter, send_request, _ = create_adapter(base_url)
    response = adapter.request("GET", url)
    assert response == send_request.return_value
    assert send_request.call_args_list == [
        call(
            HTTPRequest(
                method="GET",
                url="https://www.example.com/foo/bar",
                headers=HTTPHeaderDict(),
                body=b"",
                stream_response=False,
            )
        )
    ]


def test_no_request_url() -> None:
    adapter, _, _ = create_adapter()
    with pytest.raises(ValueError):
        adapter.request("GET")


def test_request_body_raw() -> None:
    adapter, send_request, _ = create_adapter()
    response = adapter.request("POST", "https://www.example.com", body=b"foobar")
    assert response == send_request.return_value
    assert send_request.call_args_list == [
        call(
            HTTPRequest(
                method="POST",
                url="https://www.example.com",
                headers=HTTPHeaderDict(),
                body=b"foobar",
                stream_response=False,
            )
        )
    ]


def test_request_body_json() -> None:
    adapter, send_request, _ = create_adapter()
    response = adapter.request(
        "POST",
        "https://www.example.com",
        body="foobar",
        body_encoding=HTTPBodyEncoding.JSON,
    )
    assert response == send_request.return_value
    assert send_request.call_args_list == [
        call(
            HTTPRequest(
                method="POST",
                url="https://www.example.com",
                headers=HTTPHeaderDict({"content-type": "application/json"}),
                body=b'"foobar"',
                stream_response=False,
            )
        )
    ]


def test_request_headers() -> None:
    adapter, send_request, _ = create_adapter(
        headers={
            "authorization": "Basic QWxhZGRpbjpvcGVuIHNlc2FtZQ==",
            "referer": "https://www.example.com",
        }
    )

    response = adapter.request(
        "GET",
        "https://www.example.com",
        headers={"referer": "https://www.example.com/foo/bar"},
    )
    assert response == send_request.return_value

    response = adapter.request(
        "GET",
        "https://www.example.com",
        headers={"x-foo": "bar"},
    )
    assert response == send_request.return_value

    assert send_request.call_args_list == [
        call(
            HTTPRequest(
                method="GET",
                url="https://www.example.com",
                headers=HTTPHeaderDict(
                    {
                        "authorization": "Basic QWxhZGRpbjpvcGVuIHNlc2FtZQ==",
                        "referer": "https://www.example.com/foo/bar",
                    }
                ),
                body=b"",
                stream_response=False,
            )
        ),
        call(
            HTTPRequest(
                method="GET",
                url="https://www.example.com",
                headers=HTTPHeaderDict(
                    {
                        "authorization": "Basic QWxhZGRpbjpvcGVuIHNlc2FtZQ==",
                        "referer": "https://www.example.com",
                        "x-foo": "bar",
                    }
                ),
                body=b"",
                stream_response=False,
            )
        ),
    ]


def test_request_stream_response() -> None:
    adapter, send_request, _ = create_adapter()
    response = adapter.request("GET", "https://www.example.com", stream_response=True)
    assert response == send_request.return_value
    assert send_request.call_args_list == [
        call(
            HTTPRequest(
                method="GET",
                url="https://www.example.com",
                headers=HTTPHeaderDict(),
                body=b"",
                stream_response=True,
            )
        )
    ]


def test_request_with_interceptors() -> None:
    adapter, send_request, client = create_adapter()
    adapter.request_interceptor["inter0"] = 0
    adapter.request_interceptor["inter1"] = 1
    adapter.response_interceptor["inter2"] = 1
    adapter.response_interceptor["inter3"] = 0

    response = adapter.request("GET", "https://www.example.com")

    assert client.inter0.call_args_list == [
        call(
            HTTPRequest(
                method="GET",
                url="https://www.example.com",
                headers=HTTPHeaderDict(),
                body=b"",
                stream_response=False,
            ),
            adapter,
        )
    ]
    assert client.inter1.call_args_list == [
        call(
            client.inter0.return_value,
            adapter,
        )
    ]
    assert send_request.call_args_list == [
        call(
            client.inter1.return_value,
        )
    ]
    assert client.inter3.call_args_list == [
        call(
            send_request.return_value,
            adapter,
        )
    ]
    assert client.inter2.call_args_list == [
        call(
            client.inter3.return_value,
            adapter,
        )
    ]
    assert response == client.inter2.return_value


def test_request_with_interceptors_hierarchy() -> None:
    adapter0, _, client0 = create_adapter()
    adapter1, _, client1 = create_adapter(parent_adapter=adapter0)
    adapter2, send_request2, client2 = create_adapter(parent_adapter=adapter1)

    adapter0.request_interceptor["inter0"] = 0
    adapter0.request_interceptor["inter1"] = -1
    adapter0.request_interceptor["inter2"] = -2
    adapter1.request_interceptor["inter1"] = 1
    adapter2.request_interceptor["inter2"] = 2

    response = adapter2.request("GET", "https://www.example.com")

    assert client0.inter0.call_args_list == [
        call(
            HTTPRequest(
                method="GET",
                url="https://www.example.com",
                headers=HTTPHeaderDict(),
                body=b"",
                stream_response=False,
            ),
            adapter2,
        )
    ]
    assert client1.inter1.call_args_list == [
        call(
            client0.inter0.return_value,
            adapter2,
        )
    ]
    assert client2.inter2.call_args_list == [
        call(
            client1.inter1.return_value,
            adapter2,
        )
    ]
    assert send_request2.call_args_list == [
        call(
            client2.inter2.return_value,
        )
    ]
    assert response == send_request2.return_value



import pytest

from sdkite.utils import zip_reverse


def test_zip_reverse() -> None:
    assert list(zip_reverse((1, 2), "ab")) == [(2, "b"), (1, "a")]


def test_zip_reverse_invalid_sizes() -> None:
    with pytest.raises(ValueError):
        zip_reverse((1, 2), "abc")



from sdkite.adapter import Adapter, AdapterSpec
from sdkite.client import Client
from sdkite.pagination import Pagination, paginated

try:
    from sdkite._version import __version__
except ImportError:  # pragma: no cover
    __version__ = "0.0.0.dev0-unknown"


__all__ = (
    "__version__",
    "Adapter",
    "AdapterSpec",
    "Client",
    "paginated",
    "Pagination",
)



import pytest

from sdkite import Client


class ClientA(Client):
    pass


class ClientB(Client):
    xyz: ClientA
    uvw: ClientA


class ClientC(Client):
    var_a: int  # set in init
    var_b: int  # not set in init
    var_c: ClientA  # regular case
    var_d: "ClientA"  # quoted type
    var_e: ClientB  # recursive

    def __init__(self) -> None:
        super().__init__()
        self.var_a = 42


class ClientD(Client):
    xyz: ClientC


class ClientRec0(Client):
    xyz: "ClientRec1"


class ClientRec1(Client):
    uvw: ClientRec0


def test_no_subclients() -> None:
    client = ClientA()
    assert isinstance(client, ClientA)


def test_subclients() -> None:
    client = ClientC()

    assert client.var_a == 42
    with pytest.raises(AttributeError):
        client.var_b  # pylint: disable=pointless-statement
    assert isinstance(client, ClientC)

    assert isinstance(client.var_c, ClientA)
    assert isinstance(client.var_d, ClientA)
    assert isinstance(client.var_e, ClientB)

    assert isinstance(client.var_e.xyz, ClientA)
    assert isinstance(client.var_e.uvw, ClientA)

    assert client.var_c != client.var_d
    assert client.var_e.xyz != client.var_e.uvw
    assert client.var_e.xyz != client.var_c

    # pylint: disable=protected-access
    assert client._parent is None
    assert client.var_c._parent == client
    assert client.var_d._parent == client
    assert client.var_e._parent == client
    assert client.var_e.xyz._parent == client.var_e
    assert client.var_e.uvw._parent == client.var_e


def test_subclient_custom_init() -> None:
    with pytest.raises(TypeError):
        ClientD()


def test_subclients_recursive() -> None:
    with pytest.raises(TypeError):
        ClientRec0()

    with pytest.raises(TypeError):
        ClientRec1()



from typing import TYPE_CHECKING, Any, List, cast

import pytest

from sdkite import Adapter, AdapterSpec

if TYPE_CHECKING:
    from sdkite import Client
else:
    # we want independent unit tests
    Client = object


class AdapterSimple:
    no_instance = True

    def __init__(self) -> None:
        assert AdapterSimple.no_instance
        AdapterSimple.no_instance = False


class AdapterSpecSimple(AdapterSpec[AdapterSimple]):
    def _create_adapter(self) -> AdapterSimple:
        return AdapterSimple()


class ClientSimple0(Client):
    _parent: Any = None
    adp = AdapterSpecSimple()


class ClientSimple1(Client):
    _parent: Any = None
    adp = AdapterSpecSimple()


@pytest.mark.parametrize("swap_order", [False, True])
def test_adapter_spec_simple(swap_order: bool) -> None:
    AdapterSimple.no_instance = True

    assert isinstance(ClientSimple0.adp, AdapterSpecSimple)
    assert isinstance(ClientSimple1.adp, AdapterSpecSimple)

    client0 = ClientSimple0()
    client1 = ClientSimple1()
    client1._parent = client0  # pylint: disable=protected-access

    # different order of adp instantiation
    if swap_order:
        client1.adp  # pylint: disable=pointless-statement
        client0.adp  # pylint: disable=pointless-statement
    else:
        client0.adp  # pylint: disable=pointless-statement
        client1.adp  # pylint: disable=pointless-statement

    assert isinstance(client0.adp, AdapterSimple)
    assert isinstance(client1.adp, AdapterSimple)
    assert client0.adp is client1.adp


class AdapterComplex(Adapter):
    """A complex adapter"""

    uvw: List[int]
    xyz: List[str]

    no_instance = True

    def __init__(self) -> None:
        assert AdapterComplex.no_instance
        AdapterComplex.no_instance = False
        self.var = True

    def get_attr_name(self) -> str:
        return self._attr_name

    def get_client_names(self) -> List[str]:
        return [client.__class__.__name__ for client in self._clients]

    def get_adp_attrs(self) -> List[List[str]]:
        assert self._adapters[-1] == self
        return [adp.xyz for adp in self._adapters]

    def __repr__(self) -> str:
        return f"AC<{self.uvw},{self.xyz}>"


class AdapterSpecComplex(AdapterSpec[AdapterComplex]):
    uvw = [42]

    def __init__(self, *xyz: str) -> None:
        self.xyz = list(xyz)

    def _create_adapter(self) -> AdapterComplex:
        return AdapterComplex()


class ClientComplex0(Client):
    _parent: Any = None
    adp = AdapterSpecComplex("13", "37")


class ClientComplex1(Client):
    _parent: Any = None
    adp = AdapterSpecComplex("42")


@pytest.mark.parametrize("swap_order", [False, True])
def test_adapter_spec_complex(swap_order: bool) -> None:
    AdapterComplex.no_instance = True

    assert isinstance(ClientComplex0.adp, AdapterSpecComplex)
    assert isinstance(ClientComplex1.adp, AdapterSpecComplex)
    assert ClientComplex0.adp.uvw == [42]
    assert ClientComplex1.adp.uvw == [42]
    assert ClientComplex0.adp.xyz == ["13", "37"]
    assert ClientComplex1.adp.xyz == ["42"]

    client0 = ClientComplex0()
    client1 = ClientComplex1()
    client1._parent = client0  # pylint: disable=protected-access

    # different order of adp instantiation
    if swap_order:
        client1.adp  # pylint: disable=pointless-statement
        client0.adp  # pylint: disable=pointless-statement
    else:
        client0.adp  # pylint: disable=pointless-statement
        client1.adp  # pylint: disable=pointless-statement

    assert isinstance(client0.adp, AdapterComplex)
    assert isinstance(client1.adp, AdapterComplex)

    assert client0.adp.__doc__ == "A complex adapter"
    assert repr(client0.adp) == "AC<[42],['13', '37']>"

    assert client0.adp.uvw == [42]
    assert client1.adp.uvw == [42]
    assert client0.adp.xyz == ["13", "37"]
    assert client1.adp.xyz == ["42"]
    client0.adp.uvw.append(1337)
    client0.adp.xyz.append("end")
    assert client0.adp.uvw == [42, 1337]
    assert client1.adp.uvw == [42]  # not changed
    assert client0.adp.xyz == ["13", "37", "end"]
    assert client1.adp.xyz == ["42"]  # not changed

    AdapterComplex.no_instance = True
    client0bis = ClientComplex0()
    client1bis = ClientComplex1()
    client1bis._parent = client0bis  # pylint: disable=protected-access
    assert client0bis.adp.uvw == [42]  # not changed
    assert client1bis.adp.uvw == [42]  # not changed
    assert client0bis.adp.xyz == ["13", "37"]  # # not changed not changed
    assert client1bis.adp.xyz == ["42"]  # not changed

    assert client0.adp.var is True
    assert client1.adp.var is True
    client0.adp.var = False
    assert client0.adp.var is False
    assert cast(bool, client1.adp.var) is False  # changed

    assert client0.adp.get_attr_name() == "adp"
    assert client1.adp.get_attr_name() == "adp"
    assert client0.adp.get_client_names() == ["ClientComplex0"]
    assert client1.adp.get_client_names() == ["ClientComplex0", "ClientComplex1"]
    assert client0.adp.get_adp_attrs() == [["13", "37", "end"]]
    assert client1.adp.get_adp_attrs() == [["13", "37", "end"], ["42"]]

    assert client0bis.adp.get_adp_attrs() == [["13", "37"]]
    assert client1bis.adp.get_adp_attrs() == [["13", "37"], ["42"]]


class ClientNoAdapter(Client):
    _parent: Any = None


def test_client_middle_no_adapter() -> None:
    AdapterComplex.no_instance = True

    client0 = ClientComplex0()
    client1 = ClientNoAdapter()
    client1._parent = client0  # pylint: disable=protected-access
    client2 = ClientComplex1()
    client2._parent = client1  # pylint: disable=protected-access

    client0.adp  # pylint: disable=pointless-statement
    with pytest.raises(TypeError) as excinfo:
        client2.adp  # pylint: disable=pointless-statement

    assert str(excinfo.value) == "Client 'ClientComplex1' is missing adapter spec 'adp'"


class ClientMixed1(Client):
    _parent: Any = None
    adp = AdapterSpecComplex("37")


def test_client_various_adapters() -> None:
    AdapterSimple.no_instance = True
    AdapterComplex.no_instance = True

    client0 = ClientSimple0()
    client1 = ClientMixed1()
    client1._parent = client0  # pylint: disable=protected-access

    client0.adp  # pylint: disable=pointless-statement
    with pytest.raises(TypeError) as excinfo:
        client1.adp  # pylint: disable=pointless-statement

    assert str(excinfo.value) == (
        "Client 'ClientMixed1' has a wrong adapter spec: got 'AdapterSpecComplex' "
        "instead of 'AdapterSpecSimple'"
    )



from requests import Response
from requests_mock import Mocker

from sdkite.http import HTTPHeaderDict, HTTPRequest, HTTPResponse
from sdkite.http.impl_requests import RequestsImpl, RequestsResponse


def test_requests_impl(requests_mock: Mocker) -> None:
    requests_mock.register_uri(
        # request
        "POST",
        "https://www.example.com/foo/bar",
        request_headers={"X-Foo": "Uvw"},
        # response
        content=b'{"hello":"world"}',
        reason="OK",
        headers={"X-Bar": "Xyz"},
    )

    impl = RequestsImpl()
    response = impl(
        HTTPRequest(
            method="POST",
            url="https://www.example.com/foo/bar",
            headers=HTTPHeaderDict({"X-Foo": "Uvw"}),
            body=b"hello",
            stream_response=False,
        )
    )

    request = requests_mock.request_history[0]
    assert request.scheme == "https"
    assert request.hostname == "www.example.com"
    assert request.port == 443
    assert request.path == "/foo/bar"
    assert not request.query
    assert request.text == "hello"
    assert not request.stream
    assert not request.allow_redirects

    assert isinstance(response, HTTPResponse)
    assert isinstance(response, RequestsResponse)
    assert isinstance(response.raw, Response)
    assert response.status_code == 200
    assert response.reason == "OK"
    assert response.headers == HTTPHeaderDict({"X-Bar": "Xyz"})
    assert b"".join(response.data_stream) == b'{"hello":"world"}'
    assert response.data_bytes == b'{"hello":"world"}'
    assert response.data_str == '{"hello":"world"}'
    assert response.data_json == {"hello": "world"}

    # check mutability
    response.headers["aaa"] = "bbb"
    assert response.headers == HTTPHeaderDict({"X-Bar": "Xyz", "aaa": "bbb"})
    assert not list(response.data_stream)  # exhausted
    response.data_json["aaa"] = "bbb"  # type: ignore[index]
    assert response.data_json == {"hello": "world", "aaa": "bbb"}


def test_request_impl_custom_ua(requests_mock: Mocker) -> None:
    requests_mock.register_uri(
        # request
        "GET",
        "https://www.example.com/foo/bar",
        request_headers={"User-Agent": "Custom"},
    )

    impl = RequestsImpl()
    impl(
        HTTPRequest(
            method="GET",
            url="https://www.example.com/foo/bar",
            headers=HTTPHeaderDict({"User-Agent": "Custom"}),
            body=b"",
            stream_response=False,
        )
    )



import sys
from typing import Dict, List, Optional, Set, Tuple, TypeVar
import warnings

from sdkite import Adapter, AdapterSpec
from sdkite.http.impl_requests import RequestsImpl
from sdkite.http.model import (
    HTTPBodyEncoding,
    HTTPHeaderDict,
    HTTPRequest,
    HTTPResponse,
)
from sdkite.http.utils import encode_request_body, urlsjoin
from sdkite.utils import zip_reverse

if sys.version_info < (3, 8):  # pragma: no cover
    from typing_extensions import Literal
else:  # pragma: no cover
    from typing import Literal

if sys.version_info < (3, 9):  # pragma: no cover
    from typing import Callable, Mapping
else:  # pragma: no cover
    from collections.abc import Callable, Mapping

if sys.version_info < (3, 10):  # pragma: no cover
    from typing_extensions import ParamSpec
else:  # pragma: no cover
    from typing import ParamSpec


P = ParamSpec("P")
T = TypeVar("T")


class HTTPAdapter(Adapter):
    url: Optional[str]
    headers: HTTPHeaderDict
    request_interceptor: Dict[str, int]
    response_interceptor: Dict[str, int]

    def __init__(self, send_request: Callable[[HTTPRequest], HTTPResponse]) -> None:
        self._send_request = send_request

    def request(
        self,
        method: str,
        url: Optional[str] = None,
        *,
        body: object = None,
        body_encoding: HTTPBodyEncoding = HTTPBodyEncoding.AUTO,
        headers: Optional[Mapping[str, str]] = None,
        stream_response: bool = False,
    ) -> HTTPResponse:
        #
        # create request
        #

        # method
        method = method.upper()

        # url
        url_parts = [adapter.url for adapter in self._adapters]
        if url:
            url_parts.append(url)
        url = urlsjoin(url_parts)
        if url is None:
            raise ValueError("No URL provided")

        # headers
        headers_parts: List[Optional[Mapping[str, str]]] = [
            adapter.headers for adapter in self._adapters
        ]
        headers_parts.append(headers)
        headers = HTTPHeaderDict()
        for headers_part in headers_parts:
            if headers_part:
                headers.update(headers_part)

        # body
        body, content_type = encode_request_body(body, body_encoding)
        if content_type:
            if "content-type" in headers:
                warnings.warn(
                    "The 'content-type' header is being overridden"
                    f" due to request body encoding {body_encoding}"
                    f" (from '{headers['content-type']}' to '{content_type}')",
                    RuntimeWarning,
                )
            headers["content-type"] = content_type

        # create request object
        request = HTTPRequest(
            method=method,
            url=url,
            headers=headers,
            body=body,
            stream_response=stream_response,
        )

        #
        # send request
        #

        # request interceptors
        for interceptor in self._get_interceptors("request_interceptor"):
            request = interceptor(request, self)

        # send request
        response = self._send_request(request)

        # response interceptors
        for interceptor in self._get_interceptors("response_interceptor"):
            response = interceptor(response, self)

        return response

    def _get_interceptors(
        self,
        kind: Literal["request_interceptor", "response_interceptor"],
    ) -> List[Callable[[T, "HTTPAdapter"], T]]:
        seen_names: Set[str] = set()
        interceptors: List[Tuple[int, Callable[[T, HTTPAdapter], T]]] = []
        for client, adapter in zip_reverse(self._clients, self._adapters):
            for name, order in getattr(adapter, kind).items():
                if name not in seen_names:
                    interceptors.append((order, getattr(client, name)))
                    seen_names.add(name)
        interceptors.sort(key=lambda item: (item[0], str(item[1])))
        return [interceptor for _, interceptor in interceptors]


class HTTPAdapterSpec(AdapterSpec[HTTPAdapter]):
    def __init__(
        self,
        url: Optional[str] = None,
        headers: Optional[Mapping[str, str]] = None,
    ) -> None:
        self.url = url
        self.headers = HTTPHeaderDict(headers)
        self.request_interceptor: Dict[str, int] = {}
        self.response_interceptor: Dict[str, int] = {}

    def _create_adapter(self) -> HTTPAdapter:
        # for now we force an implementation based on 'requests'
        send_request = RequestsImpl()
        return HTTPAdapter(send_request)

    def register_interceptor(
        self,
        kind: Literal["request_interceptor", "response_interceptor"],
        attr_name: str,
        order: int,
    ) -> None:
        interceptors: Dict[str, int] = getattr(self, kind)
        if attr_name in interceptors:
            warnings.warn(
                f"Interceptor '{attr_name}' of '{self._attr_name}' has already been registered"
                f" with order {interceptors[attr_name]}, ignoring new registration"
                f" with order {order}",
                RuntimeWarning,
            )
        else:
            interceptors[attr_name] = order

    def _intercept(
        self,
        kind: Literal["request_interceptor", "response_interceptor"],
        order: int,
    ) -> Callable[[Callable[P, T]], Callable[P, T]]:
        def decorator(fct: Callable[P, T]) -> Callable[P, T]:
            if sys.version_info < (3, 10):  # pragma: no cover
                if isinstance(fct, (classmethod, staticmethod)):
                    name = fct.__func__.__name__
                else:
                    name = fct.__name__
            else:  # pragma: no cover
                name = fct.__name__
            self.register_interceptor(kind, name, order)
            return fct

        return decorator

    def intercept_request(
        self, order: int
    ) -> Callable[[Callable[P, T]], Callable[P, T]]:
        return self._intercept("request_interceptor", order)

    def intercept_response(
        self, order: int
    ) -> Callable[[Callable[P, T]], Callable[P, T]]:
        return self._intercept("response_interceptor", order)



import sys

from requests import Response, Session
from urllib3.util import SKIP_HEADER  # type: ignore[attr-defined]

from sdkite.http.model import HTTPHeaderDict, HTTPRequest, HTTPResponse

if sys.version_info < (3, 8):  # pragma: no cover
    from backports.cached_property import cached_property
else:  # pragma: no cover
    from functools import cached_property

if sys.version_info < (3, 9):  # pragma: no cover
    from typing import Iterator
else:  # pragma: no cover
    from collections.abc import Iterator


class RequestsResponse(HTTPResponse):
    def __init__(self, response: Response):
        self._response = response

    @property
    def raw(self) -> Response:
        return self._response

    @property
    def status_code(self) -> int:
        return self._response.status_code

    @property
    def reason(self) -> str:
        return self._response.reason

    @cached_property
    def headers(self) -> HTTPHeaderDict:
        return HTTPHeaderDict(self._response.raw.headers)

    @cached_property
    def data_stream(self) -> Iterator[bytes]:
        return self._response.iter_content()

    @property
    def data_bytes(self) -> bytes:
        return self._response.content

    @property
    def data_str(self) -> str:
        return self._response.text

    @cached_property
    def data_json(self) -> object:
        return self._response.json()


class RequestsImpl:
    def __init__(self) -> None:
        self.session = Session()

    def __call__(self, request: HTTPRequest) -> HTTPResponse:
        headers = request.headers

        # remove request/urllib3 User-Agent header
        if "user-agent" not in headers:
            headers = HTTPHeaderDict(headers)  # copy
            headers["user-agent"] = SKIP_HEADER

        response = self.session.request(
            method=request.method,
            url=request.url,
            headers=headers,
            data=request.body,
            stream=request.stream_response,
            allow_redirects=False,
        )

        return RequestsResponse(response)
